module Schemafier
  extend ActiveSupport::Concern
  def index
    render_step_response
  end

  def show
    render_step_response
  end

  def update
    update_step_subject
  end

  def schema
    render json: get_step_subject_schema
  end
  
    # Ensure STEPS Hash keys are all symbols
  def clean_steps
    Hash[steps.map { |key, val| [key.to_sym, val] } ]
  end

  # Return all step keys (step names)
  def step_keys
    clean_steps.keys
  end

  # Return the name of the first step
  def first_step
    name, fields = clean_steps.first

    name
  end

  # Return the name of the last step
  def last_step
    name, fields = clean_steps.last

    name
  end

  # Return the name of the previous step (if past first step)
  def prev_step
    if current_step_index > 0
      step_keys[current_step_index - 1]
    end
  end

  # Return the name of the next step (if there are steps remaining)
  def next_step
    if current_step_index < step_keys.length - 1
      step_keys[current_step_index + 1]
    end
  end

  # Obtain the current step name from :id parameter in route path
  def current_step
    params.include?(:id) ? params[:id].to_sym : first_step
  end

  # Find index of the current step (lookup in step_keys)
  def current_step_index
    step_keys.index(current_step)
  end

  # Get fields for a step (default step to current_step)
  def get_step_fields(step = nil)
    step ||= current_step

    # Get schema for current model class
    schema = get_step_subject_schema

    # Get the fields for the current step
    fields = clean_steps[step.to_sym]

    # Respond with the schema for current step fields
    schema.select { |key, val| fields.include?(key.to_sym) }
  end

  # Lookup the schema of all fields for the step subject model
  def get_step_subject_schema
    # Get the model from the subject (class definition, not instance)
    model = @step_subject.class

    # Find all database fields (except id)
    fields = model.columns_hash.select { |key, value| key.to_s != "id" }

    # Return reconstructed hash of model schema
    Hash[fields.map { |key, value|
      enum_method = key.pluralize.to_sym

      if model.respond_to? enum_method
        [key, model.send(enum_method).keys]
      else
        [key, value.type]
      end
    } ]
  end

  # Render json step response (for current_step)
  def render_step_response
    response = {
      index: current_step_index,
      total: step_keys.length,
      name: current_step,
      fields: get_step_fields
    }

    response[:prev] = prev_step if prev_step
    response[:next] = next_step if next_step

    render json: response
  end

  # Update the step subject (model) with params from current_step
  def update_step_subject
    if @step_subject.update(update_params)
      ap "UPDATED!"
    else
      ap "FAILED TO UPDATE!"
      ap @step_subject
    end

    render_step_response
  end

  # Define step params permitted as only those defined in the specified step
  def update_params(step = nil)
    params.permit(get_step_fields(step).keys)
  end
end